/*
*         OpenPBS (Portable Batch System) v2.3 Software License
*
* Copyright (c) 1999-2000 Veridian Information Solutions, Inc.
* All rights reserved.
*
* ---------------------------------------------------------------------------
* For a license to use or redistribute the OpenPBS software under conditions
* other than those described below, or to purchase support for this software,
* please contact Veridian Systems, PBS Products Department ("Licensor") at:
*
*    www.OpenPBS.org  +1 650 967-4675                  sales@OpenPBS.org
*                        877 902-4PBS (US toll-free)
* ---------------------------------------------------------------------------
*
* This license covers use of the OpenPBS v2.3 software (the "Software") at
* your site or location, and, for certain users, redistribution of the
* Software to other sites and locations.  Use and redistribution of
* OpenPBS v2.3 in source and binary forms, with or without modification,
* are permitted provided that all of the following conditions are met.
* After December 31, 2001, only conditions 3-6 must be met:
*
* 1. Commercial and/or non-commercial use of the Software is permitted
*    provided a current software registration is on file at www.OpenPBS.org.
*    If use of this software contributes to a publication, product, or
*    service, proper attribution must be given; see www.OpenPBS.org/credit.html
*
* 2. Redistribution in any form is only permitted for non-commercial,
*    non-profit purposes.  There can be no charge for the Software or any
*    software incorporating the Software.  Further, there can be no
*    expectation of revenue generated as a consequence of redistributing
*    the Software.
*
* 3. Any Redistribution of source code must retain the above copyright notice
*    and the acknowledgment contained in paragraph 6, this list of conditions
*    and the disclaimer contained in paragraph 7.
*
* 4. Any Redistribution in binary form must reproduce the above copyright
*    notice and the acknowledgment contained in paragraph 6, this list of
*    conditions and the disclaimer contained in paragraph 7 in the
*    documentation and/or other materials provided with the distribution.
*
* 5. Redistributions in any form must be accompanied by information on how to
*    obtain complete source code for the OpenPBS software and any
*    modifications and/or additions to the OpenPBS software.  The source code
*    must either be included in the distribution or be available for no more
*    than the cost of distribution plus a nominal fee, and all modifications
*    and additions to the Software must be freely redistributable by any party
*    (including Licensor) without restriction.
*
* 6. All advertising materials mentioning features or use of the Software must
*    display the following acknowledgment:
*
*     "This product includes software developed by NASA Ames Research Center,
*     Lawrence Livermore National Laboratory, and Veridian Information
*     Solutions, Inc.
*     Visit www.OpenPBS.org for OpenPBS software support,
*     products, and information."
*
* 7. DISCLAIMER OF WARRANTY
*
* THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND. ANY EXPRESS
* OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
* OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT
* ARE EXPRESSLY DISCLAIMED.
*
* IN NO EVENT SHALL VERIDIAN CORPORATION, ITS AFFILIATED COMPANIES, OR THE
* U.S. GOVERNMENT OR ANY OF ITS AGENCIES BE LIABLE FOR ANY DIRECT OR INDIRECT,
* INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
* OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
* NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
* This license will be governed by the laws of the Commonwealth of Virginia,
* without reference to its choice of law rules.
*/
#ifndef _U_HASH_MAP_STRUCTS_H
#define _U_HASH_MAP_STRUCTS_H

#include "pbs_error.h"
#include "container.hpp"

/*
 * Type of data in the map
 */
/* Moved to pbs_constant.h (included by pbs_error.h) */
/* #define ENV_DATA      1 */
/* #define CMDLINE_DATA  2 */
/* #define LOGIC_DATA    3 */
/* #define STATIC_DATA   4 */
/* #define CONFIG_DATA   5 */
/* #define FILTER_DATA   6 */
/* #define SCRIPT_DATA   7 */

/* This is used for all environment data for the job
* Contents:
* all environment variables
* all parameters
* all filter information
*/
class job_data
  {
public:
    std::string name;
    std::string value;
    int var_type;
    int op_type;
    unsigned int flags;
    job_data(const char *pName,const char *pVal,int var,int op):
      var_type(var),op_type(op),flags(0)
      {
      if(pName == NULL) pName = "";
      if(pVal == NULL) pVal = "";
      name = pName;
      value = pVal;
      }
  };

typedef container::item_container<job_data *> job_data_container;
typedef job_data_container::item_iterator     job_data_iterator;

int hash_add_item(job_data_container *head, const char *name, const char *value, int var_type, int op_type);

void hash_priority_add_or_exit(job_data_container *head, const char *name, const char *value, int var_type);

void hash_add_or_exit(job_data_container *head, const char *name, const char *value, int var_type);

int hash_del_item(job_data_container *head, const char *name);

int hash_clear(job_data_container *head);

int hash_find(job_data_container *head, const char *name, job_data **env_var);

int hash_count(job_data_container *head);

int hash_print(job_data_container *head);

int hash_add_hash(job_data_container *dest, job_data_container *src, int overwrite_existing);

int hash_strlen(job_data_container *src);

#endif /* _U_HASH_MAP_STRUCTS_H */
